//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// //
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// Author: Ondrej Vondrous
// Department of Telecommunication Engineering
// Czech Technical University in Prague
// Faculty of Electrical Engineering
// 2015
//


#include "udg/applications/XMLTrafficGenerator/UDGConfig.h"
#include "inet/networklayer/common/L3AddressResolver.h"
#include "inet/networklayer/contract/ipv4/IPv4Address.h"

UDGConfig::UDGConfig(const cXMLElement *xmlFile,const inet::L3Address ip,const type_t type) {
    if (readXMLFile(xmlFile,ip,type)) {

    } else {
        exit(1);
    }
}

UDGConfig::~UDGConfig() {
    // TODO Auto-generated destructor stub
}

bool UDGConfig::readXMLFile(const cXMLElement *xmlFile, const inet::L3Address ip, const type_t type) {
    cXMLElement *xmlBuffer, *xmlProBuffer, *xmlParBuffer, *xmlDatBuffer,
            *xmlPktBuffer;
    cXMLElementList proList, parList, datList, pktList;
    cXMLElementList::iterator xmlProListIterator, xmlParListIterator,
            xmlDatListIterator;
    cXMLAttributeMap attrs;
    packet_t *tmp_pkt;
    profile_t * tmp_profile;
    queue<packet_t *> * tmp_queue;
    map<int, queue<packet_t *> *> * tmp_pkt_map;
    bool match;
    bool stop;

    xmlBuffer = NULL;
    xmlProBuffer = NULL;
    match = false;
    stop = false;

    int profile_id;

    attrs = xmlFile->getAttributes();
    if (attrs.count("ver") > 0) {
        if (attrs["ver"] != "0.1") {          //UDG profiles Version check
            throw cRuntimeError(
                    "Fatal error occurred: UDG XML :: Profile version mismatch!");
        } else {
            stringstream(attrs["ver"]) >> udg_profiles.version;
        }
    } else {
        throw cRuntimeError(
                "Fatal error occurred: Can not parse UDG XML :: Profile version!");
    }

    xmlBuffer = xmlFile->getFirstChild();
    do {
        proList = xmlBuffer->getChildren();
        xmlProListIterator = proList.begin();
        if (strcmp(xmlBuffer->getTagName(), "profile") == 0) {
            attrs = xmlBuffer->getAttributes();
            if (attrs.count("id") > 0) {
                stringstream(attrs["id"]) >> profile_id;
                tmp_profile = new profile_t;
                tmp_profile->id = profile_id;
                tmp_profile->src_ip = inet::L3Address("0.0.0.0");
                tmp_profile->dst_ip = inet::L3Address("0.0.0.0");
                tmp_profile->src_port = 0;
                tmp_profile->dst_port = 0;
                tmp_profile->stime = 0;
                for (int i = 0; xmlProListIterator != proList.end();
                        xmlProListIterator++, i++) {
                    xmlProBuffer = (*xmlProListIterator);
                    if (strcmp(xmlProBuffer->getTagName(), "params") == 0) {
                        parList = xmlProBuffer->getChildren();
                        xmlParListIterator = parList.begin();
                        for (int j = 0; xmlParListIterator != parList.end();
                                xmlParListIterator++, j++) {
                            xmlParBuffer = (*xmlParListIterator);
                            match = false;
                            if (strcmp(xmlParBuffer->getTagName(), "type")
                                    == 0) {
                                if (strcmp(xmlParBuffer->getNodeValue(),
                                        "TCP/IP") == 0) {
                                    tmp_profile->type = tcp_t;
                                    match = true;
                                } else if (strcmp(xmlParBuffer->getNodeValue(),
                                        "UDP/IP") == 0) {
                                    tmp_profile->type = udp_t;
                                    match = true;
                                } else {
                                    throw cRuntimeError(
                                            "Fatal error occurred: Can not parse UDG XML :: unknown protocol type!");
                                }
                            }
                            if (strcmp(xmlParBuffer->getTagName(), "src_ip")
                                    == 0) {
                                const char * address = xmlParBuffer->getNodeValue();
                                if (!inet::IPv4Address::isWellFormed(address)){
                                    tmp_profile->src_ip = inet::L3AddressResolver().resolve(address);
                                } else {
                                    tmp_profile->src_ip = inet::L3Address(address);
                                }
                                EV << "SRC:" << tmp_profile->src_ip << "\n";
                                match = true;
                            }
                            if (strcmp(xmlParBuffer->getTagName(), "dst_ip")
                                    == 0) {
                                const char * address = xmlParBuffer->getNodeValue();
                                if (!inet::IPv4Address::isWellFormed(address)){
                                    tmp_profile->dst_ip = inet::L3AddressResolver().resolve(address);
                                } else {
                                    tmp_profile->dst_ip = inet::L3Address(address);
                                }
                                EV << "DST:" << tmp_profile->dst_ip << "\n";
                                match = true;
                            }
                            if (strcmp(xmlParBuffer->getTagName(), "src_port")
                                    == 0) {
                                stringstream(xmlParBuffer->getNodeValue())
                                        >> tmp_profile->src_port;
                                match = true;
                            }
                            if (strcmp(xmlParBuffer->getTagName(), "dst_port")
                                    == 0) {
                                stringstream(xmlParBuffer->getNodeValue())
                                        >> tmp_profile->dst_port;
                                match = true;
                            }
                            if (!match) {
                                throw cRuntimeError(
                                        "Fatal error occurred: Can not parse UDG XML :: unknown params element!");
                            }
                        }
                        if (((tmp_profile->src_ip != ip)
                                || (tmp_profile->dst_ip != ip))
                                && (tmp_profile->type != type)) {
                            stop = true;
                        }else{
                            //initialize stime;
                            tmp_profile->stime = 0;
                        }
                    } else if ((strcmp(xmlProBuffer->getTagName(), "data")
                            && (!stop)) == 0) {
                        datList = xmlProBuffer->getChildren();
                        xmlDatListIterator = datList.begin();
                        for (int j = 0; xmlDatListIterator != datList.end();
                                xmlDatListIterator++, j++) {
                            xmlDatBuffer = (*xmlDatListIterator);
                            if (strcmp(xmlDatBuffer->getTagName(), "pkt")
                                    == 0) {
                                attrs = xmlDatBuffer->getAttributes();
                                if (attrs.count("dir") && (attrs.count("id"))) {
                                    tmp_pkt = new packet_t;
                                    tmp_pkt->prev = -1;
                                    tmp_pkt->pprev = profile_id;
                                    if (strcmp(attrs.at("dir").c_str(), "sd")
                                            == 0) {
                                        tmp_pkt->dir = sd;
                                    } else if (strcmp(attrs.at("dir").c_str(),
                                            "ds") == 0) {
                                        tmp_pkt->dir = ds;
                                    } else {
                                        throw cRuntimeError(
                                                "Fatal error occurred: Can not parse UDG XML :: unknown pkt direction!");
                                    }
                                    stringstream(attrs["id"]) >> tmp_pkt->id;

                                } else {
                                    throw cRuntimeError(
                                            "Fatal error occurred: Can not parse UDG XML :: id and dir required for pkt!");
                                }
                                EV << tmp_pkt->id << "\n";
                                if (attrs.count("pprev")
                                        || attrs.count("prev")) {
                                    if (attrs.count("pprev") > 0) {
                                        stringstream(attrs["pprev"])
                                                >> tmp_pkt->pprev;
                                        if (attrs.count("prev") == 0) {
                                            throw cRuntimeError(
                                                    "Fatal error occurred: Can not parse UDG XML :: prev pkt id required for pprev!");
                                        }
                                    } else {
                                        tmp_pkt->pprev = profile_id;
                                    }
                                    stringstream(attrs["prev"])
                                            >> tmp_pkt->prev;
                                    if (tmp_pkt->pprev == -1) {
                                        tmp_pkt->pprev = profile_id;
                                    }
                                    if (tmp_profile->reply_data.count(
                                            tmp_pkt->pprev) > 0) {
                                        tmp_pkt_map =
                                                tmp_profile->reply_data.at(
                                                        tmp_pkt->pprev);
                                    } else {
                                        tmp_pkt_map = new map<int,
                                                queue<packet_t*> *>;
                                        tmp_profile->reply_data[tmp_pkt->pprev] =
                                                tmp_pkt_map;
                                    }
                                    if ((*tmp_pkt_map)[tmp_pkt->prev] == NULL) {
                                        tmp_queue = new queue<packet_t *>;
                                        (*tmp_pkt_map)[tmp_pkt->prev] =
                                                tmp_queue;
                                    }
                                    (*tmp_pkt_map)[tmp_pkt->prev]->push(
                                            tmp_pkt);
                                } else {
                                    if (tmp_pkt->dir == sd) {
                                        tmp_profile->sd_send_data.push(tmp_pkt);
                                    } else {
                                        tmp_profile->ds_send_data.push(tmp_pkt);
                                    }
                                }
                                xmlPktBuffer = xmlDatBuffer->getFirstChild();
                                do {
                                    if (strcmp(xmlPktBuffer->getTagName(),
                                            "time") == 0) {
                                        stringstream(
                                                xmlPktBuffer->getNodeValue())
                                                >> tmp_pkt->time;
                                    } else if (strcmp(
                                            xmlPktBuffer->getTagName(),
                                            "payload") == 0) {
                                        attrs = xmlPktBuffer->getAttributes();
                                        if (attrs.count("size") > 0) {
                                            stringstream(attrs.at("size"))
                                                    >> tmp_pkt->size;
                                        } else {
                                            throw cRuntimeError(
                                                    "Fatal error occurred: Can not parse UDG XML :: size of packet is required!");
                                        }
                                        if (attrs.count("type") > 0) {
                                            tmp_pkt->type = attrs.at("type");
                                            if (strcmp(tmp_pkt->type.c_str(),
                                                    "rnd") != 0) {
                                                throw cRuntimeError(
                                                        "Fatal error occurred: Can not parse UDG XML :: unknown packet type!");
                                            }
                                        } else {
                                            tmp_pkt->type = "rnd";
                                        }
                                    } else {
                                        throw cRuntimeError(
                                                "Fatal error occurred: Can not parse UDG XML :: data element mismatch!");
                                    }
                                    xmlPktBuffer =
                                            xmlPktBuffer->getNextSibling();
                                } while (xmlPktBuffer != NULL);
                            } else {
                                throw cRuntimeError(
                                        "Fatal error occurred: Can not parse UDG XML :: unknown data element!");
                            }

                        }
                    } else {

                    }
                    tmp_pkt_map = NULL;
                    tmp_pkt = NULL;
                }

            } else {          //UDG profiles Version check
                throw cRuntimeError(
                        "Fatal error occurred: UDG XML :: Profile version mismatch!");
            }
        } else {
            throw cRuntimeError(
                    "Fatal error occurred: Can not parse UDG XML :: unexpected input!");
        }
        if (!stop) {
            udg_profiles.profiles.push_back(tmp_profile);
        }else{
            throw cRuntimeError(
                    "Fatal error occurred: UDG XML :: STOP condition occured!");
        }
        stop=false;
        xmlBuffer = xmlBuffer->getNextSibling();
    } while (xmlBuffer != NULL);
    return true;
}

vector<profile_t *> UDGConfig::getUDGProfiles(const inet::L3Address addr,
        const type_t type) {
    vector<profile_t *> tmp_vector;
    for (unsigned int i = 0; i < udg_profiles.profiles.size(); i++) {
        if (((udg_profiles.profiles[i]->src_ip==addr)
                && (udg_profiles.profiles[i]->type == type))
                || (((udg_profiles.profiles[i]->dst_ip==addr)
                        && (udg_profiles.profiles[i]->type == type)))) {
            tmp_vector.push_back(udg_profiles.profiles[i]);
        }
    }
    EV << "Loading [" << tmp_vector.size()<<"] profile(s) for host: "<< addr <<std::endl;
    std::cout << "Loading [" << tmp_vector.size()<<"] profile(s) for host: "<< addr <<std::endl;
    return tmp_vector;
}

